@@ -34,7 +34,7 @@ Galois/Counter Mode (GCM).
 
 __all__ = ['GcmMode']
 
-from Cryptodome.Util.py3compat import b, bchr, byte_string, bord, unhexlify
+from Cryptodome.Util.py3compat import byte_string, bord, unhexlify, _copy_bytes
 
 from Cryptodome.Util.number import long_to_bytes, bytes_to_long
 from Cryptodome.Hash import BLAKE2s
@@ -42,19 +42,57 @@ from Cryptodome.Random import get_random_bytes
 
 from Cryptodome.Util._raw_api import (load_pycryptodome_raw_lib, VoidPointer,
                                   create_string_buffer, get_raw_buffer,
-                                  SmartPointer, c_size_t, expect_byte_string)
+                                  SmartPointer, c_size_t, c_uint8_ptr)
 
-_raw_galois_lib = load_pycryptodome_raw_lib("Cryptodome.Util._galois",
-                    """
-                    int ghash(  uint8_t y_out[16],
-                                const uint8_t block_data[],
-                                size_t len,
-                                const uint8_t y_in[16],
-                                const void *exp_key);
-                    int ghash_expand(const uint8_t h[16],
-                                     void **ghash_tables);
-                    int ghash_destroy(void *ghash_tables);
-                    """)
+from Cryptodome.Util import _cpu_features
+
+
+# C API by module implementing GHASH
+_ghash_api_template = """
+    int ghash_%imp%(uint8_t y_out[16],
+                    const uint8_t block_data[],
+                    size_t len,
+                    const uint8_t y_in[16],
+                    const void *exp_key);
+    int ghash_expand_%imp%(const uint8_t h[16],
+                           void **ghash_tables);
+    int ghash_destroy_%imp%(void *ghash_tables);
+"""
+
+def _build_impl(lib, postfix):
+    from collections import namedtuple
+
+    funcs = ( "ghash", "ghash_expand", "ghash_destroy" )
+    GHASH_Imp = namedtuple('_GHash_Imp', funcs)
+    try:
+        imp_funcs = [ getattr(lib, x + "_" + postfix) for x in funcs ]
+    except AttributeError:      # Make sphinx stop complaining with its mocklib
+        imp_funcs = [ None ] * 3
+    params = dict(list(zip(funcs, imp_funcs)))
+    return GHASH_Imp(**params)
+
+
+def _get_ghash_portable():
+    api = _ghash_api_template.replace("%imp%", "portable")
+    lib = load_pycryptodome_raw_lib("Cryptodome.Hash._ghash_portable", api)
+    result = _build_impl(lib, "portable")
+    return result
+_ghash_portable = _get_ghash_portable()
+
+
+def _get_ghash_clmul():
+    """Return None if CLMUL implementation is not available"""
+
+    if not _cpu_features.have_clmul():
+        return None
+    try:
+        api = _ghash_api_template.replace("%imp%", "clmul")
+        lib = load_pycryptodome_raw_lib("Cryptodome.Hash._ghash_clmul", api)
+        result = _build_impl(lib, "clmul")
+    except OSError:
+        result = None
+    return result
+_ghash_clmul = _get_ghash_clmul()
 
 
 class _GHASH(object):
@@ -69,18 +107,19 @@ class _GHASH(object):
     (x^128 + x^7 + x^2 + x + 1).
     """
 
-    def __init__(self, subkey):
+    def __init__(self, subkey, ghash_c):
         assert len(subkey) == 16
 
-        expect_byte_string(subkey)
+        self.ghash_c = ghash_c
+
         self._exp_key = VoidPointer()
-        result = _raw_galois_lib.ghash_expand(subkey,
-                                              self._exp_key.address_of())
+        result = ghash_c.ghash_expand(c_uint8_ptr(subkey),
+                                      self._exp_key.address_of())
         if result:
-            raise ValueError("Error %d while expanding the GMAC key" % result)
+            raise ValueError("Error %d while expanding the GHASH key" % result)
 
         self._exp_key = SmartPointer(self._exp_key.get(),
-                                     _raw_galois_lib.ghash_destroy)
+                                     ghash_c.ghash_destroy)
 
         # create_string_buffer always returns a string of zeroes
         self._last_y = create_string_buffer(16)
@@ -88,14 +127,13 @@ class _GHASH(object):
     def update(self, block_data):
         assert len(block_data) % 16 == 0
 
-        expect_byte_string(block_data)
-        result = _raw_galois_lib.ghash(self._last_y,
-                                       block_data,
-                                       c_size_t(len(block_data)),
-                                       self._last_y,
-                                       self._exp_key.get())
+        result = self.ghash_c.ghash(self._last_y,
+                                    c_uint8_ptr(block_data),
+                                    c_size_t(len(block_data)),
+                                    self._last_y,
+                                    self._exp_key.get())
         if result:
-            raise ValueError("Error %d while updating GMAC" % result)
+            raise ValueError("Error %d while updating GHASH" % result)
 
         return self
 
@@ -106,6 +144,7 @@ class _GHASH(object):
 def enum(**enums):
     return type('Enum', (), enums)
 
+
 MacStatus = enum(PROCESSING_AUTH_DATA=1, PROCESSING_CIPHERTEXT=2)
 
 
@@ -134,7 +173,7 @@ class GcmMode(object):
     :undocumented: __init__
     """
 
-    def __init__(self, factory, key, nonce, mac_len, cipher_params):
+    def __init__(self, factory, key, nonce, mac_len, cipher_params, ghash_c):
 
         self.block_size = factory.block_size
         if self.block_size != 16:
@@ -143,14 +182,19 @@ class GcmMode(object):
 
         if len(nonce) == 0:
             raise ValueError("Nonce cannot be empty")
-        if not byte_string(nonce):
+        if isinstance(nonce, str):
             raise TypeError("Nonce must be a byte string")
 
-        self.nonce = nonce
+        # See NIST SP 800 38D, 5.2.1.1
+        if len(nonce) > 2**64 - 1:
+            raise ValueError("Nonce exceeds maximum length")
+
+
+        self.nonce = _copy_bytes(None, None, nonce)
         """Nonce"""
 
         self._factory = factory
-        self._key = key
+        self._key = _copy_bytes(None, None, key)
         self._tag = None  # Cache for MAC tag
 
         self._mac_len = mac_len
@@ -174,39 +218,41 @@ class GcmMode(object):
         hash_subkey = factory.new(key,
                                   self._factory.MODE_ECB,
                                   **cipher_params
-                                  ).encrypt(bchr(0) * 16)
+                                  ).encrypt(b'\x00' * 16)
 
         # Step 2 - Compute J0 (integer, not byte string!)
-        if len(nonce) == 12:
-            self._j0 = bytes_to_long(nonce + b("\x00\x00\x00\x01"))
+        if len(self.nonce) == 12:
+            self._j0 = bytes_to_long(self.nonce + b"\x00\x00\x00\x01")
         else:
             fill = (16 - (len(nonce) % 16)) % 16 + 8
-            ghash_in = (nonce +
-                        bchr(0) * fill +
+            ghash_in = (self.nonce +
+                        b'\x00' * fill +
                         long_to_bytes(8 * len(nonce), 8))
-            self._j0 = bytes_to_long(_GHASH(hash_subkey)
+            self._j0 = bytes_to_long(_GHASH(hash_subkey, ghash_c)
                                      .update(ghash_in)
                                      .digest())
 
         # Step 3 - Prepare GCTR cipher for encryption/decryption
+        nonce_ctr = long_to_bytes(self._j0 >> 32, 12)
+        iv_ctr = (self._j0 + 1) & 0xFFFFFFFF
         self._cipher = factory.new(key,
                                    self._factory.MODE_CTR,
-                                   initial_value=self._j0 + 1,
-                                   nonce=b(""),
+                                   initial_value=iv_ctr,
+                                   nonce=nonce_ctr,
                                    **cipher_params)
 
         # Step 5 - Bootstrat GHASH
-        self._signer = _GHASH(hash_subkey)
+        self._signer = _GHASH(hash_subkey, ghash_c)
 
         # Step 6 - Prepare GCTR cipher for GMAC
         self._tag_cipher = factory.new(key,
                                        self._factory.MODE_CTR,
                                        initial_value=self._j0,
-                                       nonce=b(""),
+                                       nonce=b"",
                                        **cipher_params)
 
         # Cache for data to authenticate
-        self._cache = b("")
+        self._cache = b""
 
         self._status = MacStatus.PROCESSING_AUTH_DATA
 
@@ -229,7 +275,7 @@ class GcmMode(object):
         invoke this method multiple times, each time with the next segment.
 
         :Parameters:
-          assoc_data : byte string
+          assoc_data : bytes/bytearray/memoryview
             A piece of associated data. There are no restrictions on its size.
         """
 
@@ -242,6 +288,11 @@ class GcmMode(object):
 
         self._update(assoc_data)
         self._auth_len += len(assoc_data)
+
+        # See NIST SP 800 38D, 5.2.1.1
+        if self._auth_len > 2**64 - 1:
+            raise ValueError("Additional Authenticated Data exceeds maximum length")
+
         return self
 
     def _update(self, data):
@@ -249,7 +300,7 @@ class GcmMode(object):
 
         if len(self._cache) > 0:
             filler = min(16 - len(self._cache), len(data))
-            self._cache += data[:filler]
+            self._cache += _copy_bytes(None, filler, data)
             data = data[filler:]
 
             if len(self._cache) < 16:
@@ -257,10 +308,10 @@ class GcmMode(object):
 
             # The cache is exactly one block
             self._signer.update(self._cache)
-            self._cache = b("")
+            self._cache = b""
 
         update_len = len(data) // 16 * 16
-        self._cache = data[update_len:]
+        self._cache = _copy_bytes(update_len, None, data)
         if update_len > 0:
             self._signer.update(data[:update_len])
 
@@ -275,7 +326,7 @@ class GcmMode(object):
         #   See step 6 in section 7.2
         len_cache = len(self._cache)
         if len_cache > 0:
-            self._update(bchr(0) * (16 - len_cache))
+            self._update(b'\x00' * (16 - len_cache))
 
     def encrypt(self, plaintext):
         """Encrypt data with the key and the parameters set at initialization.
@@ -298,7 +349,7 @@ class GcmMode(object):
         This function does not add any padding to the plaintext.
 
         :Parameters:
-          plaintext : byte string
+          plaintext : bytes/bytearray/memoryview
             The piece of data to encrypt.
             It can be of any length.
         :Return:
@@ -320,6 +371,10 @@ class GcmMode(object):
         self._update(ciphertext)
         self._msg_len += len(plaintext)
 
+        # See NIST SP 800 38D, 5.2.1.1
+        if self._msg_len > 2**39 - 256:
+            raise ValueError("Plaintext exceeds maximum length")
+
         return ciphertext
 
     def decrypt(self, ciphertext):
@@ -343,7 +398,7 @@ class GcmMode(object):
         This function does not remove any padding from the plaintext.
 
         :Parameters:
-          ciphertext : byte string
+          ciphertext : bytes/bytearray/memoryview
             The piece of data to decrypt.
             It can be of any length.
 
@@ -418,7 +473,7 @@ class GcmMode(object):
         tampered with while in transit.
 
         :Parameters:
-          received_mac_tag : byte string
+          received_mac_tag : bytes/bytearray/memoryview
             This is the *binary* MAC, as received from the sender.
         :Raises ValueError:
             if the MAC does not match. The message has been tampered with
@@ -459,7 +514,7 @@ class GcmMode(object):
         """Perform encrypt() and digest() in one step.
 
         :Parameters:
-          plaintext : byte string
+          plaintext : bytes/bytearray/memoryview
             The piece of data to encrypt.
         :Return:
             a tuple with two byte strings:
@@ -474,7 +529,7 @@ class GcmMode(object):
         """Perform decrypt() and verify() in one step.
 
         :Parameters:
-          ciphertext : byte string
+          ciphertext : bytes/bytearray/memoryview
             The piece of data to decrypt.
           received_mac_tag : byte string
             This is the *binary* MAC, as received from the sender.
@@ -500,12 +555,12 @@ def _create_gcm_cipher(factory, **kwargs):
         GCM has been only defined for `Cryptodome.Cipher.AES`.
 
     :Keywords:
-      key : byte string
+      key : bytes/bytearray/memoryview
         The secret key to use in the symmetric cipher.
         It must be 16 (e.g. *AES-128*), 24 (e.g. *AES-192*)
         or 32 (e.g. *AES-256*) bytes long.
 
-      nonce : byte string
+      nonce : bytes/bytearray/memoryview
         A value that must never be reused for any other encryption.
 
         There are no restrictions on its length,
@@ -532,4 +587,11 @@ def _create_gcm_cipher(factory, **kwargs):
         nonce = get_random_bytes(16)
     mac_len = kwargs.pop("mac_len", 16)
 
-    return GcmMode(factory, key, nonce, mac_len, kwargs)
+    # Not documented - only used for testing
+    use_clmul = kwargs.pop("use_clmul", True)
+    if use_clmul and _ghash_clmul:
+        ghash_c = _ghash_clmul
+    else:
+        ghash_c = _ghash_portable
+
+    return GcmMode(factory, key, nonce, mac_len, kwargs, ghash_c)
